from wikifile.wikiFile import WikiFile
from wikifile.smw import SMWPart
from wikifile.metamodel import Property
import json
import logging
import sys
import jinja2
from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
import re

from lodstorage.jsonable import JSONAble, Types


class Toolbox(object):
    """Bundles methods that are required by the commandline tools that this file provides"""

    def __init__(self):
        pass
        
    def getParser(self):
        # Setup argument parser
        self.parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter)
        self.parser.add_argument('-p', '--pages', dest="pages",  nargs='+',
                            help="Names of the pages the action should be applied to")
        self.parser.add_argument('--BackupPath', dest="backupPath", help="Path to store/update the wiki entries",
                            required=True)
        self.parser.add_argument('-stdin', dest="stdin", action='store_true',
                            help='Use the input from STD IN using pipes')
        self.parser.add_argument('-ex', '--exclude', dest="exclude_keys",
                            help="List of keys that should be excluded")
        self.parser.add_argument('--debug', dest="debug", action='store_true', default=False, help="Enable debug mode")

    def fromJson(self, listname: str, jsonStr: str, sample: list = None):
        """
        Convert the given string to JSON and convert the values to the corresponding datatype"
        Args:
            listname: name of the list the data is stored in
            jsonStr: data as json string that should be converted
            sample: sample data of the actual data. Used to convert the values to the correct type
        Returns:
            
        """""
        types = Types(listname)
        types.getTypes("data", sample, 1)
        json_parsed = JSONAble()
        json_parsed.fromJson(jsonStr, types)
        return json_parsed.__dict__[listname]

    @classmethod
    def wikiSontoLOD(cls, wiki_sample, entity="Event"):
        regex = '{{ *%s(?:.|\r|\n)*?\}}' % entity
        re_groups = re.search(regex, wiki_sample)
        if re_groups is not None:
            property_list = re_groups.group().replace('}}', '').split('|')[1:]
            wikidict = {}
            for i in property_list:
                mapping = i.strip().split('=')
                try:
                    wikidict[mapping[0].strip()] = int(mapping[1].strip())
                except:
                    wikidict[mapping[0].strip()] = mapping[1].strip()
            return wikidict
        return None

    @classmethod
    def dicttoWikiSon(self, dic, entity="Event"):
        wikiSon = "{{%s\n" % entity
        for key in dic:
            if len(str(dic[key]).strip()) != 0:
                wikiSon += "|%s = %s \n" % (str(key), str(dic[key]))
        wikiSon += "}}"
        return wikiSon


import io

from wikifile.wikiFile import WikiFile
from wikibot.wikipush import WikiPush
import json
import logging
import os
import csv
import sys
from wikifile.toolbox import Toolbox
from wikifile.wikiRender import WikiRender


class WikiFix(Toolbox):

    def __init__(self, fromWikiId, debug=False):
        super(WikiFix, self).__init__()
        self.wikiPush = WikiPush(fromWikiId=fromWikiId, login=True)
        self.debug = debug
        self.wikiRender = WikiRender()

    def getEventsinSeries(self, seriesName, queryField):
        """
        Get all the events given an event series and query label
        ToDo: This method is to specific for wikirender we better migrate it to openresearch -> e.g get all pageTitels of events of given series result is than passed to wikirender which handels the extraction

        Args:
            seriesName(str): Name of the series
            queryField(str): Query field used in SMW ask to get pages of subseries
        Returns:
            pageTitles(list): list of all pageTitles found in given series
        """
        query = '[[%s::%s]]' % (queryField, seriesName)
        pageTitles = self.wikiPush.query(query)
        return list(pageTitles)

    def get_Properties(self, pageDicts):
        '''
        gets the Header of the CSV
        Args:
            pageDicts: Page dictionaries
        Returns:
            Header(set): Properties found in the Events to append to CSV as a header
        '''
        if type(pageDicts) != list:
            return set(pageDicts.keys())
        else:
            header = set()
            for page in pageDicts:
                header.update(set(page.keys()))
            return header

    def exportToCsv(self, header, pageDicts, filename='dict', propertyList=None):
        '''
        Args:
            header(list): Header for the CSV File
            pageDicts(List of Dics): List of Dicts of pages
            filename(str)(optional): name of the csv file
        Returns:
            CSV file with filename
        '''
        if propertyList is not None:
            header = propertyList
        with open(filename + '.csv', 'w', newline='') as csv_file:
            writer = csv.writer(csv_file)
            writer.writerow(header)
            for page in pageDicts:
                pagelist = []
                for field in header:
                    checker = page.get(field)
                    if checker is None:
                        pagelist.append('')
                    else:
                        pagelist.append(checker)
                writer.writerow(pagelist)

    def editProperty(self, pageDict, propertyName, propertyValue):
        """
        Edit the given property in the Dict
        Args:
            pageDict(dict): Dictionary which holds the dictionary
            propertyName(str): Property to change/add
            propertyValue: Property Value to set
        Returns:
            pageDict(dict):updated page dict
        """
        pageDict[propertyName] = propertyValue
        return pageDict

    def getCSVofPages(self, pageTitles: str, wikiSonName: str):
        """
        Extracts for the given page titles the given WikiSon object and returns it as csv
        Args:
            pageTitles:
            wikiSonName:

        Returns:

        """
        pass

    def getCsv(self, pageTitles):
        '''
        converts the given pages to csv.
        Args:
            pageTitles: Page names to fetch and convert to CSV
        Returns:
            header(list): Header for CSV File
            pageDicts(List of Dics): List of Dicts of pages
        '''
        pageDicts = []
        for page in pageTitles:
            pageItem = self.wikiPush.fromWiki.getPage(page)
            wikiFile = WikiFile("sampleFile", "tmp", self.wikiRender, pageItem.text())
            pageDict = wikiFile.extract_template("Event")
            if self.debug:
                print(pageDict)
            pageDict['pageTitle'] = page
            pageDicts.append(pageDict)
            header = self.get_Properties(pageDicts)
        return header, pageDicts

    def exportWikiSonToCSVFile(self, filename: str, pageTitels: list, wikiSonName: str, pageTitleKey: str = "pageTitle",
                           properties: list = [], limitProperties: bool = False) -> str:
        """
        Exports the given given WikiSon entities corresponding to the given WikiSonName of the pages corresponding to
        the given pageTitles and returns the values as csv.

        Args:
            filename(str)(optional): name of the csv file with path
            pageTitles(list): List of all pageTitles from which the given WikiSon entity should be extracted
            wikiSonName(str): Name of the WikiSon object that should be extracted
            pageTitleKey(str): Name of the key that should be used to identify the pageTitle. This name should be distinct form other properties of the object
            properties(list): List of property names that should occur in the returned LoD. Order of the list is used as order of result. Default is null.
            limitProperties(bool): If true the resulting dicts only contain keys that are present in the given properties list any other key is removed. Otherwise all properties that are either given or defined are present in the result. Defualt is False.

        Returns:
            List of dicts containing the WikiSon entities of the given pages
        """
        csvString=self.exportWikiSonToCSV(pageTitels, wikiSonName, pageTitleKey, properties, limitProperties)
        with open(filename, 'w', newline='') as file:
            file.write(csvString)

    def exportWikiSonToCSV(self, pageTitels: list, wikiSonName: str, pageTitleKey: str = "pageTitle",
                           properties: list = [], limitProperties: bool = False) -> str:
        """
        Exports the given given WikiSon entities corresponding to the given WikiSonName of the pages corresponding to
        the given pageTitles and returns the values as csv.

        Args:
            pageTitles(list): List of all pageTitles from which the given WikiSon entity should be extracted
            wikiSonName(str): Name of the WikiSon object that should be extracted
            pageTitleKey(str): Name of the key that should be used to identify the pageTitle. This name should be distinct form other properties of the object
            properties(list): List of property names that should occur in the returned LoD. Order of the list is used as order of result. Default is null.
            limitProperties(bool): If true the resulting dicts only contain keys that are present in the given properties list any other key is removed. Otherwise all properties that are either given or defined are present in the result. Defualt is False.

        Returns:
            List of dicts containing the WikiSon entities of the given pages
        """
        lod = self.exportWikiSonToLOD(pageTitels, wikiSonName, pageTitleKey, properties, limitProperties)
        keys = lod[0].keys()  # exportWikiSonToLOD ensures that all returned recorde have the same set of keys
        csvString = ""
        csvStream=io.StringIO(csvString)
        dict_writer = csv.DictWriter(csvStream, keys)
        dict_writer.writeheader()
        dict_writer.writerows(lod)
        csvStream.seek(0)
        csvString=csvStream.read()
        return csvString

    def exportWikiSonToLOD(self, pageTitels: list, wikiSonName: str, pageTitleKey: str = "pageTitle",
                           properties: list = [], limitProperties: bool = False) -> list:
        """
        Exports the given given WikiSon entities corresponding to the given WikiSonName of the pages corresponding to
        the given pageTitles and returns the values as list of dicts.

        Args:
            pageTitles(list): List of all pageTitles from which the given WikiSon entity should be extracted
            wikiSonName(str): Name of the WikiSon object that should be extracted
            pageTitleKey(str): Name of the key that should be used to identify the pageTitle. This name should be distinct form other properties of the object
            properties(list): List of property names that should occur in the returned LoD. Order of the list is used as order of result. Default is null.
            limitProperties(bool): If true the resulting dicts only contain keys that are present in the given properties list any other key is removed. Otherwise all properties that are either given or defined are present in the result. Defualt is False.

        Returns:
            List of dicts containing the WikiSon entities of the given pages
        """
        lod = []
        for pageTitle in pageTitels:
            wikiFile = self.getWikiFile(pageTitle)
            wikiSon = wikiFile.extract_template(wikiSonName)
            if wikiSon is not None:
                wikiSon[pageTitleKey] = pageTitle
                lod.append(wikiSon)

        # Build up the set of keys
        keys = set()
        if not limitProperties:
            for record in lod:
                used_keys = {key for key in record.keys()}
                keys.update(used_keys)

        # Add key to the properties
        for key in keys:
            if key not in properties:
                properties.append(key)

        # clean up the entries by adding missing keys
        for record in lod:
            for key in properties:
                if key not in record:
                    record[key] = None
            if limitProperties:
                recordKeys=list(record.keys())
                for key in recordKeys:
                    if key not in properties:
                        del record[key]

        # order dicts according to the order of the property list
        propertyMap = {key: pos for pos, key in enumerate(properties)}
        for pos, record in enumerate(lod):
            lod[pos] = dict(sorted(record.items(), key=lambda x: propertyMap[x[0]]))
        return lod

    def convertLODToCSV(self, pageTitles: list, wikiSon="Event"):
        """
        Converts the given LOD to csv by placing the key inside the values with the given pageTitleKey as column
        Example:
            Input:
        Args:
            pageTitles:
            wikiSon:

        Returns:
            header(list): Header for CSV File
            pageDicts(List of Dics): List of Dicts of pages
        """
        # ToDo: Ensure that the headers and values align when converting the LOD to csv
        pageDicts = []
        for pageTitle in pageTitles:
            wikiFile = self.getWikiFile(pageTitle)
            pageDict = wikiFile.extract_template(wikiSon)
            if self.debug:
                print(pageDict)
            pageDict['pageTitle'] = pageTitle
            pageDicts.append(pageDict)
            header = self.get_Properties(pageDicts)
        return header, pageDict

    def convertWikiFilesToLOD(self, wikiFiles: list, wikiSonName: str):
        '''
        converts the given pagesTitles to csv by extracting the given templateName from the wikiPage corresponding to
        the given pageTitle.

        Args:
            wikiFiles(list): PageTitles to fetch and convert to CSV
            wikiSonName(str): Name of the WikiSon object that should be extracted

        Returns:
            header(list): Header for CSV File
            pageDicts(List of Dics): List of Dicts of pages
        '''
        lod = {}
        for wikifile in wikiFiles:
            if isinstance(wikifile, WikiFile):
                values = wikifile.extract_template(wikiSonName)
                if values is None:
                    values = {}
                pageTitle = wikifile.getPageTitle()
                if pageTitle is not None:
                    lod[pageTitle] = values
        return lod

    def exportCsvToWiki(self, pageContentList):
        """
            Creates a WikiSon Template of the given CSV
            Args:
                pageContentList(dict): List of Dicts of pageTitle and content of page in Wikison Format(1 row = 1 page = 1 entry)
            Returns:
                listOfFailures(str): List containing pageTitles of the pages failed to upload to wiki
        """
        listofFailures = []
        for pageContent in pageContentList:
            pageTitle = list(pageContent.keys())[0]
            try:
                page = self.wikiPush.fromWiki.getPage(pageTitle)
                page.edit(pageContent[pageTitle],
                          f"modified through csv import by {self.wikiPush.fromWiki.wikiUser.user}")
                if self.debug:
                    print('Successfully pushed the page to the wiki')
            except Exception as ex:
                listofFailures.append(pageTitle)
                if self.debug:
                    print('Error pushing to wiki')
                    print(ex)
        return listofFailures

    def prepareExportCsvContent(self, filename):
        """
        Creates a WikiSon Template of the given CSV
        Args:
            filename(str):CSV File name
        Returns:
            pageContentList(dict): List of Dicts of pagename and content of page in Wikison Format(1 row = 1 page = 1 entry)
        """
        pageContentList = []
        with open(filename, 'r') as csv_file:
            reader = csv.DictReader(csv_file)
            for row in reader:
                pageTitle = row['pageTitle']
                del row['pageTitle']
                pageText = Toolbox.dicttoWikiSon(row)
                pairDict = dict()
                pairDict[pageTitle] = pageText
                pageContentList.append(pairDict)
        return pageContentList

    def importCSVFileToWiki(self, csvFileName: str, wikiSonName: str, pageTitle: str = "pageTitle"):
        """
        Imports the given csv file to the wiki by applying the values of the rows.
        It is assumed that each row is clearly identified by one column which represents a pageTitle in the wiki
        Args:
            csvString: csv content as string
            wikiSonName(str): Name of the wikiSon object that should be updated/created
            pageTitle(str): Column name that holds the pageTitle

        Returns:
            Nothing
        """
        csvContent = self.readFile(filename=csvFileName)
        return self.importCSVContentToWiki(csvContent, wikiSonName=wikiSonName, pageTitle=pageTitle)

    def importCSVContentToWiki(self, csvString: str, wikiSonName: str, pageTitle: str = "pageTitle"):
        """
        Imports the given csv content to the wiki by applying the values of the rows.
        It is assumed that each row is clearly identified by one column which represents a pageTitle in the wiki
        Args:
            csvString: csv content as string
            wikiSonName(str): Name of the wikiSon object that should be updated/created
            pageTitle(str): Column name that holds the pageTitle

        Returns:
            Nothing
        """
        lod = self.convertCSVtoLOD(csvString, pageTitle)
        self.importLODtoWiki(lod, wikiSon=wikiSonName, titleKey=pageTitle)

    def readFile(self, filename: str) -> str:
        """
        Reads the given filename and returns it as string
        Args:
            filename: Name of the file that should be returned as string

        Returns:
            Content of the file as string
        """
        content = ""
        with open(filename, 'r') as file:
            content = file.read()
        return content

    def convertCSVtoLOD(self, csvString: str, pageTitle: str = "pageTitle"):
        """
        Converts the given csv string to a list of dicts (LOD).
        It is assumed that each row is clearly identified by one column which represents a pageTitle in the wiki
        Args:
            csvString(str): csv content as string
            pageTitle(str): Column name that holds the pageTitle

        Returns:
            List of dicts where each key represents the pageTitle and the value the WIKISon values corresponding to the page
        """
        reader = csv.DictReader(io.StringIO(csvString))
        return list(reader)

    def importLODtoWiki(self, data: list, wikiSon: str, titleKey: str = "pageTitle"):
        """
        Uses the given data and updates the corresponding pages in the wiki.
        It is assumed that the page title is part of the dict binding the values of the dict to the corresponding page.

        Args:
            data(list): List of Dicts that should be used to update the corresponding pages in the wiki
            wikiSon(str): Name of the wikiSon object that should be updated/created
            titleKey(str): Name of the key that holds the pageTitle. Default is "pageTitle"

        Returns:

        """
        pageDict = self.pagesListtoDict(data, titleKey)
        wiki_files = self.getUpdatedPages(pageDict, wikiSon)
        self.pushWikiFilesToWiki(wiki_files)

    def pagesListtoDict(self, data: list, titleKey: str = "pageTitle") -> dict:
        """
        Converts the given list of dicts to a dict in which each value is identified by the corresponding pageTitle.
        It is assumed that each dict has the pageTitle key set and that this key is unique within the list.
        Example:
            Input: [{"pageTitle":"Test Page", "label":Test}]
            Output: {"Test Page":{"label":Test}}

        Args:
            data(list): List of dicts of WikiSon values
            titleKey(str): Name of the key that holds the pageTitle. Default is "pageTitle"

        Returns:
            dict of dicts where each dict is identified by the pageTitle
        """
        res = {}
        for record in data:
            if titleKey in record:
                pageTitle = record[titleKey]
                del record[titleKey]
                res[pageTitle] = record
        return res

    def pushWikiFilesToWiki(self, wiki_files: list):
        """
        Pushes the content of the given wikiFiles to the corresponding wiki pages in the wiki
        
        Args:
            wiki_files: list of WikiFiles that should be pushed to the wiki

        Returns:
            Nothing
        """
        for wiki_file in wiki_files:
            if isinstance(wiki_file, WikiFile):
                page_content = str(wiki_file)
                update_msg = f"modified through csv import by {self.wikiPush.fromWiki.wikiUser.user}"
                wiki_file.getPage().edit(page_content, update_msg)

    def getUpdatedPages(self, records: dict, wikiSon: str) -> list:
        """
        Updates the wikiPages with the given values by applying the values to the WikiSon object matching the given wikiSon name.
        The update is applied by creating a WikiFile object from the page contnent in the wiki.
        The changes are then applied in the WikiFile object.
        Note: At this point the changes are not applied to the wiki. Do do so use pushWikiFilesToWiki()

        Args:
            records(dict): dict of dicts in which the new values for a page are identified by the pageTitle
            wikiSon(str): Name of the wikiSon object that should be updated/created

        Returns:
            List of WikiFile objects with the updated content
        """
        res = []
        for pageTitle, values in records.items():
            if pageTitle is not None:
                wiki_file = self.getUpdatedPage(pageTitle, values, wikiSon)
                res.append(wiki_file)
        return res

    def getUpdatedPage(self, pageTitle: str, values: dict, wikiSon: str) -> WikiFile:
        """
        Updates the given page with the given values

        Args:
            pageTitle(str): Title of the page that shoud be updated
            values(dict): values that should be used for the update
            wikiSon(str): Name of the WikiSon that should be updated

        Returns:
            WikiFile corresponding to the given pageTitle with the applied updates
        """
        wiki_file = self.getWikiFile(pageTitle)
        wiki_file.add_template(wikiSon, values, overwrite=True)
        return wiki_file

    def getWikiFile(self, pageTitle: str) -> WikiFile:
        """
        Queries the given page and converts it to a WikiFile.

        Args:
            pageTitle: Title of the page that should be retrieved

        Returns:
            WikiFile corresponding the the given pageTitle
        """
        pageItem = self.wikiPush.fromWiki.getPage(pageTitle)
        wiki_file = WikiFile(pageTitle, "tmp", self.wikiRender, pageItem.text())
        wiki_file.setPage(pageItem)
        return wiki_file


if __name__ == "__main__":
    fix = WikiFix('ormk', debug=True)
    # pageTitles= fix.getEventsinSeries('3DUI', 'Event in series')
    # header,dicts= fix.getCsv(pageTitles)
    # fix.exportToCsv(header,dicts)
    pageContentList = fix.prepareExportCsvContent('dict.csv')
    fix.exportCsvToWiki(pageContentList)



'''
Created on 2021-03-21

@author: th
'''
import json
import re
import sys

from lodstorage.jsonable import Types, JSONAble
from wikifile.smw import SMW, Query
import pkgutil

class MetaModelElement(object):
    '''
    a generic MetaModelElement
    
    to handle the technicalities of being a MetaModelElement so that derive
    MetaModelElements can focus on the MetaModel domain specific aspects
    '''

    @classmethod
    def get_samples(cls):
        samples = [{"name": "Element"}]
        return samples

    def __init__(self, propList: list = None, properties: dict = None, template=None):
        '''
        construct me from the given properties and the given propertyMap
        
        Args:
            propList(list): a list of field names
            properties(dict): a dictionary of properties
            template(str): the template to be used
            lenient(bool): if False throw an exception if invalid property names are in the properties dict
        '''
        super().__init__()
        self.template=template
        if propList is None:
            propList = MetaModelElement.get_prop_list_from_samples(MetaModelElement.get_samples())
        self.propList=propList
        self.fromProperties(properties)

    def fromProperties(self, properties: dict):
        '''
        initialize me from the given properties
        Args:
            properties(dict): a dictionary of properties
        '''
        if properties is None:
            properties = {}
        propMap={}
        for propName in self.propList:
            # https://stackoverflow.com/a/1176023/1497139
            # snakeName= MetaModelElement.get_snake_name(propName)
            propMap[propName]=propName
        for key in properties.keys():
            if key in propMap:
                name=propMap[key]
                value=properties[key]
                self.__dict__[name]=value
        for propName in propMap.keys():
            fieldname=propMap[propName]
            if fieldname not in self.__dict__:
                self.__dict__[fieldname]=None

    @staticmethod
    def get_snake_name(name):
        # ToDo: Check if this method is still needed after migrating to camelCase property names
        return re.sub(r'(?<!^)(?=[A-Z])', '_', name).lower()


    @classmethod
    def fromJson(self, listname: str, jsonStr: str, sample: list = None):
        """
        Convert the given string to JSON and convert the values to the corresponding datatype"
        Args:
            listname: name of the list the data is stored in
            jsonStr: data as json string that should be converted
            sample: sample data of the actual data. Used to convert the values to the correct type
        Returns:

        """""
        types = Types(listname)
        types.getTypes("data", sample, 1)
        json_parsed = JSONAble()
        json_parsed.fromJson(jsonStr, types)
        return json_parsed.__dict__[listname]

    @staticmethod
    def get_parameters(attr: dict, prop_list: list):
        """
        Removes all key value pairs for which the key is not in the given prop_list.
        Args:
            attr: Dict of attributes with values that should be used for the conversion
            prop_list: List of properties that should be converted
        Returns:
            Returns list of dict containing the property names from the given prop_list with the corresponding value from attr
        """
        para_map = {}
        for para_item in [{prop: attr[prop]} for prop in prop_list]:
            para_map.update(para_item)
        return para_map

    @staticmethod
    def get_prop_list_from_samples(samples: list):
        """Returns a list of used keys by the given list of dicts"""
        prop_list = []
        for sample in samples:
            for key in sample.keys():
                if key not in prop_list:
                    prop_list.append(key)
        return prop_list

    def get_pageTitle(self, prefix:str, withNamespace=True):
        """
        Returns page name of this metamodel element. If pageTitle is not defined, it is assumed that name corresponds to the pageTitle.
        Args:
            prefix: MetaModelElement prefix e.g. properties usually have the prefix Property
            withNamespace: If true also the namespace of this property is returned. Otherwise, without namespace.
        Returns:
            Returns page name of this metamodel element.
        """
        if self.pageTitle:
            if withNamespace:
                return self.pageTitle
            else:
                return self.pageTitle.replace(f"{prefix}:", "")
        elif self.name:
            if withNamespace:
                return f"{prefix}:{self.name}"
            else:
                return self.name
        else:
            return None


class Context(MetaModelElement):
    """
    A Context groups some topics like a Namespace/Package.
    This class provides helper functions and constants to render a Context to corresponding wiki pages
    """
    samples = [{
        "name": "MetaModel"
    }]
    propList = MetaModelElement.get_prop_list_from_samples(samples)
    
    def __init__(self,properties: dict=None):
        super(Context,self).__init__(Context.propList,properties,)


class Topic(MetaModelElement):
    """
    Provides helper functions and constants for Topics
    """
    samples = [{"pageTitle": "Concept:Topic",
                "name": "Topic",
                "pluralName": "Topics",
                "icon": "File:Topic_icon.png",
                "iconUrl": "/images/a/ae/Index.png",
                "documentation": "A Topic is a Concept/Class/Thing/Entity",
                "wikiDocumentation": "A Topic is a Concept/Class/Thing/Entity",
                "context": "MetaModel",
                "listLimit": 7,
                "cargo": True
                }]
    # propList contains all attributes that belong to the Topic domain (needed to distinguish them from technical attributes)
    propList = MetaModelElement.get_prop_list_from_samples(samples)

    @classmethod
    def get_samples(cls):
        return Topic.samples

    def __init__(self, topic_properties: dict, properties: list=None):
        """
        Initialize topic with the given parameters
        Args:
            topic_properties: properties defining/describing the topic
            properties: properties that the topic uses
        """
        super(Topic, self).__init__(Topic.propList, properties=topic_properties)
        self.properties = properties

    def get_pageTitle(self, withNamespace=True):
        """
        Returns page name of this topic. If pageTitle is not defined, it is assumed that name corresponds to the pageTitle.
        Args:
            withNamespace: If true also the namespace of this property is returned. Otherwise, without namespace.
        Returns:
            Returns page name of this topic.
        """
        return super(Topic, self).get_pageTitle("Concept", withNamespace)


    def get_related_topics(self, properties):
        """
        Returns a list of names of topics that are related to this topic.
        Args:
            properties: List of all properties of the wiki

        Returns:
            list of names of related topics
        """
        res = set([])
        for p in UML.get_outgoing_edges(self.name, properties):
            # source (self) --> target
            res.add(p['target'])
        for p in UML.get_incoming_edges(self.name, properties):
            # source --> target (self)
            res.add(p['source'])
        if self.name in res:
            res.remove(self.name)
        return res

    def render_entity(self, oneliner=True):
        """Render tis topic to its template representation"""
        separator = "" if oneliner else "\n"
        res = "{{Topic" + separator
        res += SMW.render_parameters(oneliner, **MetaModelElement.get_parameters(self.__dict__, self.propList))
        return res + "}}"

    def get_page_link(self, withLabel=True, withDescription=False):
        """Returns a link to this page"""
        pageTitle = self.get_pageTitle()
        if withDescription:
            pageTitle += "::@@@"
        if withLabel:
            return f"[[{pageTitle}|{self.name}]]"
        else:
            return f"[[Concept:{pageTitle}]]"

    @classmethod
    def from_wiki_json(cls, topic_json, prop_json=None):
        """
        Parse given topic_json to attributes of this Topic and convert the values to proper python types.
        If prop_json is also given it will be converted to a list of Properties of this topic
        Args:
            topic_json: json string containing all information about the topic that should be created
            prop_json: json string containing all properties of this topic

        Returns:
            Topic object based on the given parameters
        """
        key = "data"
        topic_props = MetaModelElement.fromJson(key, topic_json, Topic.get_samples())
        properties = None
        if prop_json:
            properties = [Property(x) for x in MetaModelElement.fromJson("data", prop_json, Property.get_samples())]
        return Topic(topic_props[0], properties)

    def query_topic_overview(self, oneliner=False):
        """Returns a SMW query querying a brief description over this topic"""
        query = Query(mainlabel="-").select("Topic name::" + self.name)
        query.printout("Topic icon", "icon")
        query.printout("", "Topic")
        query.printout("Topic pluralName", "pluralName")
        query.printout("Topic documentation", "documentation")
        return query.render(oneliner=oneliner)

    def query_entities(self, limit=200, order="ascending", oneliner=False, withProperties=True, **kwargs):
        """Returns a SMW query that queries entities of this topic"""
        mainlabel = None
        if withProperties:
            mainlabel = self.name
        query = Query(mainlabel=mainlabel, limit=limit, order=order, **kwargs)
        query.select("Concept:" + self.name)
        if withProperties:
            query.printout_list_of_properties(Property.get_entity_properties(self.name, self.properties))
        return query.render(oneliner=oneliner)

    def query_examples(self, limit=10, oneliner=False, **kwargs):
        """Returns a SMW query that queries some examples of this topic entity"""
        return self.query_entities(limit=limit, order=None, searchlabel="...more examples", withProperties=False, oneliner=True, **kwargs)
        #return Query(limit=limit, searchlabel="...more examples").select("Concept:" + self.name).render(oneliner)

    def query_entity_properties(self, format="wikitable", oneliner=False):
        """Returns a SMW query querying all properties of this entity"""
        query = Query(format=format)
        query.select("Concept:Property")
        query.select("Property topic::Concept:" + self.name)
        query.printout_list_of_properties(Property.get_property_properties())
        return query.render(oneliner=oneliner)

    def query_number_entities(self, oneliner=True):
        """Returns a SMW query querying the number of entities of this topic in the wiki"""
        return Query(format="count").select("Concept:" + self.name).render(oneliner)

    def query_documentation(self, **kwargs):
        """Returns a SMW query querying the documentation of this topic"""
        query = Query(mainlabel="-", **kwargs)
        query.select("Topic name::" + self.name)
        query.printout("Topic wikiDocumentation", " ")
        return query.render()


class Property(MetaModelElement):
    """
    Provides helper functions and constants for properties
    """
    samples = [{"name": "Title",
                "label": "Offical Name",
                "type": "Special:Types/Test",
                "index": 2,
                "sortPos": 2,
                "primaryKey": False,
                "mandatory": True,
                "namespace": "Test",
                "size": 25,
                "uploadable": False,
                "defaultValue": "Some Title",
                "inputType": "combobox",
                "allowedVaues": "Tile A, Title B",
                "documentation": " Documentation can contain\n line breaks",
                "values_from": "property=Title",
                "showInGrid": False,
                "isLink": False,
                "nullable": False,
                "topic": "Concept:Event"
                },
               # Properties that are not included in the MetaModel
               {"placeholder": "e.g. SMWCon",
                "regexp": "NaturalNumber",
                "usedFor": "Concept:Event, Concept:Event series",
                "pageTitle": "Property:Title"
                }]
    @classmethod
    def get_samples(cls):
        """Returns a list of property samples as list of dicts."""
        return Property.samples

    # propList contains all attributes that belong to the Property domain / Needed to distinguish them from technical attributes
    propList = MetaModelElement.get_prop_list_from_samples(samples)

    def __init__(self, properties: dict=None):
        """Initialize property with the given properties as attributes"""
        super(Property, self).__init__(Property.propList, properties)

    def is_used_for(self):
        """Returns the topics this property is used for/by"""
        res = []
        if self.topic:
            if isinstance(self.topic, list):
                res.extend(self.topic)
            else:
                res.append(self.topic)
        if self.usedFor:
            if isinstance(self.usedFor, list):
                res.extend(self.usedFor)
            else:
                res.append(self.usedFor)
        return res

    def get_pageTitle(self, withNamespace=True):
        """
        Returns page name of this property. If pageTitle is not defined, it is assumed that name corresponds to the pageTitle.
        Args:
            withNamespace: If true also the namespace of this property is returned. Otherwise, without namespace.
        Returns:
            Returns page name of this property.
        """
        return super(Property, self).get_pageTitle(self.__class__.__name__, withNamespace)


    @staticmethod
    def get_property_properties():
        """Returns the properties that describe properties"""
        properties_json = json.loads(pkgutil.get_data("wikifile", "/resources/metamodel/Property_properties.json"))
        properties = [Property(x) for x in properties_json]
        # Add pageTitle
        for property in properties:
            property.__dict__["pageTitle"] = f"Property:Property {property.name}"
        print([p.pageTitle for p in properties])
        return properties

    @staticmethod
    def get_entity_properties(entity_name: str, properties: list):
        """
        Extracts the properties of an entity form the given property list
        Args:
            entity_name: Name of the property for which the properties should be extracted
            properties: List of all properties

        Returns:
            List of all properties the are used by the given entity
        """
        res = []
        for property in properties:
            if f"Concept:{entity_name}" in property.is_used_for():
                res.append(property)
        return sorted(res, key=lambda k: int(k.index) if k.index is not None else sys.maxsize) # ToDo: Add sortPos to all properties. If done exchange index with sortPos

    @staticmethod
    def get_primitive_properties(entity_name: str, properties: list):
        """
        Returns a list of all primitive datatype of the given entity.
        Args:
            entity_name: name of the entity for which the primitive properties should be returned
            properties: list of all properties
        Returns:
            List of primitive properties of the given entity
        """
        primitive_datatypes = ["Special:Types/Boolean",
                               "Special:Types/Date",
                               "Special:Types/Code",
                               "Special:Types/Text",
                               "Special:Types/text",
                               "Special:Types/url",
                               "Special:Types/URL",
                               "Special:Types/External identifier",
                               "Special:Types/ExternalIdentifier",
                               "Special:Types/Geographic coordinate",
                               "Special:Types/Number",
                               "Special:Types/URI",
                               "Special:Types/Url",
                               "Special:Types/number",
                               "Special:Types/code"]
        primitive_properties = []
        for property in properties:
            if f"Concept:{entity_name}" in property.is_used_for() and property.type in primitive_datatypes:
                primitive_properties.append(property)
        return primitive_properties

    def render_entity(self, oneliner=True):
        """Render this property to its template representation"""
        separator = "" if oneliner else "\n"
        res = "{{Property" + separator
        res += SMW.render_parameters(oneliner=oneliner, **MetaModelElement.get_parameters(self.__dict__, self.propList))
        return res + "}}"

    def get_page_link(self, name=None):
        """
        Returns a page link to this property. If a name is provided this name will be used instead
        Args:
            name: Name of the property page. If not given pageTitle of this object will be used.
        Returns:
            SMW page link to this property
        """
        if name is None:
            name = self.get_pageTitle(withNamespace=True)
        if self.label:
            return f"[[{name}|{self.label}]]"
        else:
            return f"[[{name}]]"

    def get_description_page_link(self):
        """
        Returns a page link to this property with the addition that a wiki language dependent description is shown if
        defined.
        For more details see https://www.semantic-mediawiki.org/wiki/Help:Special_property_Has_property_description
        Returns:
            SMW page link to this property
        """
        name = f"{self.get_pageTitle(withNamespace=False)}::@@@"
        return self.get_page_link(name)


    @staticmethod
    def filterLoP(lop: list, key: str, match: list, invert=False):
        """
        Filter given list of dicts by key value match
        Args:
            lop: list of dicts that should be filtered
            key: key to look for the match
            match: list of allowed values for the given key
            invert: If True all dicts that do NOT match are returned. Otherwise all dicts that match are returned
        Returns:
            List of dicts that have the given key with a matching value
        """
        if invert:
            return [p for p in lop if p.__dict__[key] not in match]
        else:
            return [p for p in lop if p.__dict__[key] in match]

    @classmethod
    def from_wiki_json(cls, prop_json):
        """
        Parse given prop_json to attributes of this Property and convert the values to proper python types
        Args:
            prop_json: json containing the property properties
        Returns:
            Property object
        """
        key = "data"
        prop_props = MetaModelElement.fromJson(key, prop_json, Property.get_samples())
        return Property(prop_props[0])



class UML:
    """
    ToDo: Migrate the functionality of this class to Topic
    Provides helper functions for uml generation
    """

    @staticmethod
    def get_outgoing_edges(entity_name: str, properties: list):
        """
        Reduce the given list of properties to those properties which have the given entity as subject
        Args:
            entity_name: Name of the entity for which the outgoing edges should be extracted
            properties: List of all properties
        Returns:
            List of dicts with each dict containing information for one edge.
        Examples: For entity_name=Project
            [{
            "target": "Goal",
            "source": "Project",
            "source_cardinality": "*",
            "target_cardinality": "1",
            "properties": <Property object>
            }]
            Means in UML:
            "Project" "*" --> "1" "Goal" : "<Property object>.name"
        """
        outgoing_edges = []
        for property in properties:
            if f"Concept:{entity_name}" in property.is_used_for() and property.type == "Special:Types/Page" and property.values_from:
                uml_infos = {}
                uml_infos["source"] = entity_name
                if property.values_from:
                    values_from = str(property.values_from).split("=")
                    if values_from[0] == "concept":
                        uml_infos["target"] = values_from[1]
                    else:
                        continue
                if property.inputType == "tokens":
                    uml_infos["target_cardinality"] = "*"
                else:
                    uml_infos["target_cardinality"] = "1"
                uml_infos["source_cardinality"] = "*"
                uml_infos["property"] = property
                outgoing_edges.append(uml_infos)
        return outgoing_edges

    @staticmethod
    def get_incoming_edges(entity_name: str, properties: list):
        """
        Reduce the given list of properties to those properties which have the given entity as object
        Args:
            entity_name: Name of the entity for which the incoming edges should be extracted
            properties: List of all properties
        Returns:
            List of dicts with each dict containing information for one edge.
        Examples: For entity_name=Goal
            [{
            "target": "Goal",
            "source": "Project",
            "source_cardinality": "*",
            "target_cardinality": "1",
            "properties": <Property object>
            }]
            Means in UML:
            "Project" "*" --> "1" "Goal" : "<Property object>.name"
        """
        incoming_edges = []
        for property in properties:
            if property.values_from == f"concept={entity_name}" and property.type == "Special:Types/Page" and property.values_from:
                uml_infos = {}
                uml_infos["target"] = entity_name
                if property.inputType == "tokens":
                    uml_infos["target_cardinality"] = "*"
                else:
                    uml_infos["target_cardinality"] = "1"
                uml_infos["source_cardinality"] = "*"
                uml_infos["property"] = property
                for source in property.is_used_for():
                    temp_uml_infos = uml_infos.copy()
                    temp_uml_infos["source"] = source.replace("Concept:","")
                    incoming_edges.append(temp_uml_infos)
        return incoming_edges

    @staticmethod
    def get_incoming_edges_reduced(entity_name: str, properties: list):
        """
        Same as get_incoming_edges but merges the edges that come from teh same source.
        Returns a list of dict. The edges lname is stored in the kex properties.
        Args:
            entity_name: Name of the entity for which the incoming edges should be extracted
            properties: List of all properties
        Returns:
            List of dicts with each dict containing information for one edge.
        Examples:
            [{
            "target": "Goal",
            "source": "Project",
            "source_cardinality": "*",
            "target_cardinality": "*",
            "properties": ["Task goals", "Project goals"]
            }]
            Means in UML:
            "Project" "*" --> "*" "Goal" : "Task goals, Project goals"
        """
        incoming_edges  = UML.get_incoming_edges(entity_name, properties)
        res = []
        for edge in incoming_edges:
            if edge.get('source') == entity_name:
                continue
            if edge.get('source') in [x.get('source') for x in res]:
                # merge edges
                for x in res:
                    if x.get('source') == edge.get('source'):
                        x.get('properties').append(edge.get('property').name)
            else:
                edge['properties'] = [edge.get('property').name]
                res.append(edge)
        return res

